﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.ServiceModel;
using HeuristicLab.Services.Hive.Common;
using HeuristicLab.Services.Hive.Common.DataTransfer;
using HeuristicLab.Services.Hive.Common.ServiceContracts;

namespace HeuristicLab.Services.Hive {

  /// <summary>
  /// Implementation of the Hive service (interface <see cref="IHiveService"/>).
  /// </summary>
  [ServiceBehavior(InstanceContextMode = InstanceContextMode.PerCall)]
  public class HiveService : IHiveService {
    private DataAccess.IHiveDao dao {
      get { return ServiceLocator.Instance.HiveDao; }
    }
    private HeuristicLab.Services.Hive.DataAccess.TransactionManager trans {
      get { return ServiceLocator.Instance.TransactionManager; }
    }
    private IAuthorizationManager auth {
      get { return ServiceLocator.Instance.AuthorizationManager; }
    }
    private ILifecycleManager lifecycleManager {
      get { return ServiceLocator.Instance.LifecycleManager; }
    }
    private HeartbeatManager heartbeatManager {
      get { return ServiceLocator.Instance.HeartbeatManager; }
    }

    #region Job Methods
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public Guid AddJob(Job job, JobData jobData, IEnumerable<Guid> resourceIds) {
      using (trans.OpenTransaction()) {
        job.SetState(JobState.Waiting, auth.UserId);
        job.Id = dao.AddJob(job);
        jobData.JobId = job.Id;
        jobData.LastUpdate = DateTime.Now;
        if (resourceIds != null) {
          foreach (Guid slaveGroupId in resourceIds) {
            dao.AssignJobToResource(job.Id, slaveGroupId);
          }
        } else {
          // todo: use default group
        }
        dao.AddJobData(jobData);
        return jobData.JobId;
      }
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public Guid AddChildJob(Guid parentJobId, Job job, JobData jobData) {
      using (trans.OpenTransaction()) {
        job.ParentJobId = parentJobId;
        return AddJob(job, jobData, dao.GetAssignedResources(parentJobId).Select(x => x.Id));
      }
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public Job GetJob(Guid jobId) {
      return dao.GetJob(jobId);
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public IEnumerable<Job> GetJobs() {
      return dao.GetJobs(x => true);
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public IEnumerable<LightweightJob> GetLightweightJobs(IEnumerable<Guid> jobIds) {
      return dao.GetJobs(x => jobIds.Contains(x.JobId)).Select(x => new LightweightJob(x)).ToArray();
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public IEnumerable<LightweightJob> GetLightweightChildJobs(Guid? parentJobId, bool recursive, bool includeParent) {
      return GetChildJobs(parentJobId, recursive, includeParent).Select(x => new LightweightJob(x)).ToArray();
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public JobData GetJobData(Guid jobId) {
      return dao.GetJobData(jobId);
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public void UpdateJob(Job job) {
      using (trans.OpenTransaction()) {
        dao.UpdateJob(job);
      }
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public void UpdateJobData(Job job, JobData jobData) {
      using (trans.OpenTransaction()) {
        jobData.LastUpdate = DateTime.Now;
        dao.UpdateJob(job);
        dao.UpdateJobData(jobData);
      }
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public void DeleteJob(Guid jobId) {
      using (trans.OpenTransaction()) {
        dao.DeleteJob(jobId);
      }
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public void DeleteChildJobs(Guid parentJobId) {
      using (trans.OpenTransaction()) {
        var jobs = GetChildJobs(parentJobId, true, false);
        foreach (var job in jobs) {
          dao.DeleteJob(job.Id);
          dao.DeleteJobData(job.Id);
        };
      }
    }

    #endregion

    #region Job Control Methods
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public void StopJob(Guid jobId) {
      using (trans.OpenTransaction()) {
        throw new NotImplementedException();
      }
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public void PauseJob(Guid jobId) {
      using (trans.OpenTransaction()) {
        throw new NotImplementedException();
      }
    }
    #endregion

    #region HiveExperiment Methods
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public HiveExperiment GetHiveExperiment(Guid id) {
      return dao.GetHiveExperiments(x =>
             x.HiveExperimentId == id
          && (x.OwnerUserId == auth.UserId || x.HiveExperimentPermissions.Count(hep => hep.Permission != Permission.NotAllowed && hep.GrantedUserId == auth.UserId) > 0)
          ).FirstOrDefault();
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public IEnumerable<HiveExperiment> GetHiveExperiments() {
      return dao.GetHiveExperiments(x => x.OwnerUserId == auth.UserId || x.HiveExperimentPermissions.Count(hep => hep.Permission != Permission.NotAllowed && hep.GrantedUserId == auth.UserId) > 0);
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public Guid AddHiveExperiment(HiveExperiment hiveExperimentDto) {
      using (trans.OpenTransaction()) {
        hiveExperimentDto.OwnerUserId = auth.UserId;
        hiveExperimentDto.DateCreated = DateTime.Now;
        return dao.AddHiveExperiment(hiveExperimentDto);
      }
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public void UpdateHiveExperiment(HiveExperiment hiveExperimentDto) {
      using (trans.OpenTransaction()) {
        dao.UpdateHiveExperiment(hiveExperimentDto);
      }
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public void DeleteHiveExperiment(Guid hiveExperimentId) {
      using (trans.OpenTransaction()) {
        HiveExperiment he = dao.GetHiveExperiment(hiveExperimentId);
        dao.DeleteHiveExperiment(hiveExperimentId); // child jobs will be deleted by db-trigger
      }
    }
    #endregion

    #region Login Methods
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public void Hello(Slave slaveInfo) {
      using (trans.OpenTransaction()) {
        var slave = dao.GetSlave(slaveInfo.Id);

        if (slave == null) {
          dao.AddSlave(slaveInfo);
        } else {
          dao.UpdateSlave(slaveInfo);
        }
      }
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public void GoodBye(Guid slaveId) {
      using (trans.OpenTransaction()) {
        var slave = dao.GetSlave(slaveId);
        if (slave != null) {
          slave.SlaveState = SlaveState.Offline;
          dao.UpdateSlave(slave);
        }
      }
    }
    #endregion

    #region Heartbeat Methods
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public List<MessageContainer> Heartbeat(Heartbeat heartbeat) {
      using (trans.OpenTransaction()) {
        return heartbeatManager.ProcessHeartbeat(heartbeat);
      }
    }
    #endregion

    #region Plugin Methods
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public Guid AddPlugin(Plugin plugin, List<PluginData> pluginDatas) {
      using (trans.OpenTransaction()) {
        plugin.UserId = auth.UserId;
        plugin.DateCreated = DateTime.Now;
        Guid pluginId = dao.AddPlugin(plugin);
        foreach (PluginData pluginData in pluginDatas) {
          pluginData.PluginId = pluginId;
          dao.AddPluginData(pluginData);
        }
        return pluginId;
      }
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public IEnumerable<Plugin> GetPlugins() {
      return dao.GetPlugins(x => true);
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public IEnumerable<PluginData> GetPluginDatas(List<Guid> pluginIds) {
      List<PluginData> pluginDatas = new List<PluginData>();

      using (trans.OpenTransaction()) {
        foreach (Guid guid in pluginIds) {
          List<PluginData> pluginData = dao.GetPluginDatas(x => x.PluginId == guid).ToList();
          if (pluginData != null) {
            pluginDatas.AddRange(pluginData);
          } else {
            //ignore ? 
          }
        }
        return pluginDatas;
      }
    }

    #endregion

    #region Slave Methods
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public Guid AddSlave(Slave slave) {
      using (trans.OpenTransaction()) {
        return dao.AddSlave(slave);
      }
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public Guid AddSlaveGroup(SlaveGroup slaveGroup) {
      using (trans.OpenTransaction()) {
        return dao.AddSlaveGroup(slaveGroup);
      }
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public Slave GetSlave(Guid slaveId) {
      return dao.GetSlave(slaveId);
    }

    public SlaveGroup GetSlaveGroup(Guid slaveGroupId) {
      return dao.GetSlaveGroup(slaveGroupId);
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public IEnumerable<Slave> GetSlaves() {
      return dao.GetSlaves(x => true);
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public IEnumerable<SlaveGroup> GetSlaveGroups() {
      return dao.GetSlaveGroups(x => true);
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public void UpdateSlave(Slave slave) {
      using (trans.OpenTransaction()) {
        dao.UpdateSlave(slave);
      }
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public void UpdateSlaveGroup(SlaveGroup slaveGroup) {
      using (trans.OpenTransaction()) {
        dao.UpdateSlaveGroup(slaveGroup);
      }
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public void DeleteSlave(Guid slaveId) {
      using (trans.OpenTransaction()) {
        dao.DeleteSlave(slaveId);
      }
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public void DeleteSlaveGroup(Guid slaveGroupId) {
      using (trans.OpenTransaction()) {
        dao.DeleteSlaveGroup(slaveGroupId);
      }
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public void AddResourceToGroup(Guid slaveGroupId, Guid resourceId) {
      using (trans.OpenTransaction()) {
        var resource = dao.GetResource(resourceId);
        resource.ParentResourceId = slaveGroupId;
        dao.UpdateResource(resource);
      }
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public void RemoveResourceFromGroup(Guid slaveGroupId, Guid resourceId) {
      using (trans.OpenTransaction()) {
        var resource = dao.GetResource(resourceId);
        resource.ParentResourceId = null;
        dao.UpdateResource(resource);
      }
    }

    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    // [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Client)]
    public Guid GetResourceId(string resourceName) {
      using (trans.OpenTransaction()) {
        var resource = dao.GetResources(x => x.Name == resourceName).FirstOrDefault();
        if (resource != null) {
          return resource.Id;
        } else {
          return Guid.Empty;
        }
      }
    }
    #endregion

    #region Helper Methods
    private IEnumerable<Job> GetChildJobs(Guid? parentJobId, bool recursive, bool includeParent) {
      var jobs = new List<Job>(dao.GetJobs(x => parentJobId == null ? !x.ParentJobId.HasValue : x.ParentJobId.Value == parentJobId));

      if (includeParent) {
        jobs.Add(GetJob(parentJobId.Value));
      }

      if (recursive) {
        var childs = new List<Job>();
        foreach (var job in jobs) {
          childs.AddRange(GetChildJobs(job.Id, recursive, false));
        }
        jobs.AddRange(childs);
      }
      return jobs;
    }

    #endregion
  }
}
